#! /usr/bin/wish
#############################################################################
#
# TTTTTTTT        fvri.tcl
#    TT           Author(s): Bill Sherman
#    TT           Created: May 5, 2005  (ie. 5/5/5)
#    TT           Last Modified: June 20, 2006
#   TTTT
#
# FreeVR Remote Interface
#
# A Tcl/TK script that connects to a running FreeVR application and
#   allows the operator to view and manipulate particular settings
#   of the application.
#
# Copyright 2021, Bill Sherman & Friends, All rights reserved.
# With the intent to provide an open-source license to be named later.
#############################################################################
#
# USAGE:
#	% fvri.tcl
#	[NOTE: there are currently no arguments, so the program assumes
#	the FreeVR application has a "telnet" process with a socket open
#	on port 3000 of the localhost.]
#
# HISTORY:
#	5-9 May 2005 (Bill Sherman) -- initial testing.  Worked to just
#		make the socket connection, and then do a finite loop to
#		read process information and display it in the TK window.
#
#	13-20 June 2006 (Bill Sherman) -- major improvements to convert
#		this to a usable interface to a running FreeVR application.
#
# TODO:
#	Arguments to the script that allow the host and port to be specified.
#
#	PARTIAL: Cleanup the TK window
#
#	Popup a warning/explanation window when a socket connection cannot
#		be made -- ie. explain how to check for a running telnet
#		process that (for the moment) is listening on port 3000
#
#	Popup a message window when a socket disconnect signal is recieved
#		NOTE: this needs to wait until we get to the point where we
#		are in a continual loop.
#
#	Add the "$appstatus" information to the top section, and add a
#		corresponding checkbox to keep it updated in the loop
#
#	Add a toggle for stereo rendering (actually will just be an adjustment
#		to the users' IOD value -- or perhaps a global IOD setting)
#
#	Add a setting for the line-rendering mode.
#
#	Provide an option to view a list of all the inputs (probably a popup)
#
#	Add new information about the running system:
#		[Information from vrContextInfo]
#		- appstatus
#		- sysstatus
#		- startup result
#		- paused flag
#		- paused_time
#		- shared memory usage amount
#
#	Add the ability to close the current connection, and a button to
#		"reconnect" (ie. connect to whatever might now be running).
#		This will then allow me to run this program once, and
#		continue to connect to subsequant VR programs.
#
#	Add the ability to have a text window (scrollable) into which FreeVR
#		telnet commands can be typed, with the response placed into
#		the text window.
#
#	Consider adding widgets to control various inputs, in particular the
#		6-sensors, so for example a wand might be moved 1 ft up per
#		click on a particular button, or rotate 10deg around X.
#		This could be tied into the input-list frame.
#		[Request from Phil McDonald.]
#
#	Show all the users, not just User[0].
#
#############################################################################


###################################################################
## Some functions that I borrowed from various other sources.

###############
## dialog_create{}: from "Effective Tcl/Tk Programming" section 6.4, page 239
proc dialog_create {class {win "auto"}} {
	if {$win == "auto"} {
		set count 0
		set win ".dialog[incr count]"
		while {[winfo exists $win]} {
			set win ".dialog[incr count]"
		}
	}

	toplevel $win -class $class
	frame $win.info
	pack $win.info -expand yes -fill both -padx 2 -pady 2

	frame $win.sep -height 2 -borderwidth 1 -relief sunken
	pack $win.sep -fill x -pady 4

	frame $win.controls
	pack $win.controls -fill x -padx 4 -pady 4

	wm title $win $class
	wm group $win .

	after idle [format {
		update idletasks
		wm minsize %s [winfo reqwidth %s] [winfo reqheight %s]
	} $win $win $win]

	return $win
}

###############
## dialog_info{} & dialog_controls{} are from ET/TP, section 6.4, page 240
## NOTE: it's not really clear to me what these two functions buy us.  I'll
##   probably delete them in a near-future version.
proc dialog_info {win} {
	return "$win.info"
}
proc dialog_controls {win} {
	return "$win.controls"
}

###############
## notice_show{} is also from ET/TP, section 6.4, page 240-241
option add *Notice*dismiss.text "Dismiss" widgetDefault
proc notice_show {mesg {icon "info"}} {
	set top [dialog_create Notice]
	set x [expr [winfo rootx .] + 50]
	set y [expr [winfo rooty .] + 50]
	wm geometry $top "+$x+$y"

	set info [dialog_info $top]
	label $info.icon -bitmap $icon
	pack $info.icon -side left -padx 8 -pady 8
	label $info.mesg -text $mesg -wraplength 8i -justify left
	pack $info.mesg -side right -expand yes -fill both -padx 8 -pady 8

	set cntls [dialog_controls $top]
	button $cntls.dismiss -command "destroy $top"
	pack $cntls.dismiss -pady 4

	return $top
}

###############
## char{} is from http://foto.hut.fi/~jussi/tcl/Fragments.html
# NOTE: I swapped which line is commented out, and renamed from "chr" to "char"
proc char c { 
#   if {[string length $c] > 1 } { error "char: arg should be a single char"}
    set c [string range $c 0 0] 
    set v 0; 
    scan $c %c v; return $v
}    
###################################################################


###################################################################
##### Globals
# Even though I'm generally opposed to the use of global variables,
#   there are some exceptions, and working in scripting languages
#   is often one of them, expecially when just developing a crude
#   widget interface to something.
# Here is the list of globals:
set stoploop_flag 0
set num_procs 0
set num_windows 0
set num_users 0


###################################################################
## SendCommandGetInput{} sends a command to the socket, and waits
##   for the input -- returning the input
proc SendCommandGetInput {command} {
	global sock

	puts $sock $command
	flush $sock
	after 1;		# Need to wait since we're expecting input

	set result -1
	while {$result == -1} {
		set result [gets $sock input]
	}

#puts "For command '$command' Got input '$input' with result '$result'"

	return $input
}

###################################################################
## SendCommandNoInput{} sends a command to the socket, but does not
##   expect a response, so does not get the input
proc SendCommandNoInput {command} {
	global sock

	puts $sock $command
	flush $sock
	after 5;
}


###################################################################
## StopLoop{} sets a global flag to indicate the user has clicked on
##   the exit button
proc StopLoop {} {
	global stoploop_flag
	set stoploop_flag 1
	puts "set flag for exiting"
	puts $stoploop_flag
}


###################################################################
## HelpPopup{} pops up a message about how to use this program
proc HelpPopup {} {
	set mesg [format "\
How to use the FreeVR Remote Interface: ...
"]
	notice_show $mesg
}


###################################################################
## AboutPopup{} pops up a message about this program
proc AboutPopup {} {
	set mesg [format "\
FreeVR Remote Interface\n\
By: Bill Sherman\n\
Last Modified: 06/15/2006\n\
\n\
A simple Tcl/Tk program to interface with a running FreeVR application\
"]
	notice_show $mesg
}


###################################################################
## UIPopup{} pops up a message about how to use the FreeVR application
proc UIPopup {} {
	global sock

	puts "about to get the ui information"
	## NOTE: we probably should clear the incoming buffer of all
	##   messages at this point, though if there are any, then that
	##   indicates a problem elsewhere
	set mesg [format "The application reports the following information about its usage:\n"]

	puts $sock "print ui"
	flush $sock
	after 5;		# Need to wait since we're expecting input
	gets $sock input
	while {$input != "==uiend=="} {
#puts [format "UIPopup: just got '%s' string" $input]
		set mesg [format "%s\n%s" $mesg $input]
		gets $sock input
	}

	notice_show $mesg
}


###################################################################
##### SystemUpdate{} function reads the basic system information, and
# fills in the appropriate TK values.  (NOTE: not intended for looping.)
#
# NOTE: I do a wait ("after") command after sending commands, in order
#   to give time for a reponse.  A better solution would be to actually
#   loop until a response is recieved.  But, since I should probably
#   also make this a procedure, I'll wait to solve that issue until that
#   point.
proc SystemUpdate {} {
	global num_procs num_windows
	global syslist

	## Get the application name
	set input [SendCommandGetInput "echo \$name"]
	.top.appname configure -text [format "Application: %s" $input]

	## Get the application authors
	set input [SendCommandGetInput "echo \$authors"]
	.top.authname configure -text [format "Application author(s): %s" $input]

	## Get the library information
	set input [SendCommandGetInput "echo \$version \[\$target\]"]
	.top.libname configure -text [format "Library: %s" $input]


	## Now get the actual system info
	set input [SendCommandGetInput "p system"]
	set syslist [split $input :]
	set num_procs [lindex $syslist 1]
	.top.sysname configure -text [format "System: %s -- %d processes" [lindex $syslist 0] $num_procs]

	#for { set count 2 } { $count < [expr $num_procs + 2] } { incr count } {
	#	.top.proclist insert end [lindex $syslist $count]
	#}


	## Now get the number (and list) of windows
	set input [SendCommandGetInput "p windows"]

	set num_windows [lindex [split $input :] 0]
	puts [format "There are %d windows\n" $num_windows]


	## Now get the number (and list) of users
	set input [SendCommandGetInput "p users"]

	set num_users [lindex [split $input :] 0]
	puts [format "There are %d users\n" $num_users]
}


###################################################################
##### ToggleAllWindows{} function 
proc ToggleAllWindows { setting value } {
	global num_windows

	puts [format "ToggleAllWindows: setting = %s, value = %d\n" $setting $value]

	set command_string [format "set window\[%%d\] %s %d" $setting $value]
	puts $command_string

	for { set count 0 } { $count < $num_windows } { incr count } {
		set command [format $command_string $count]
		SendCommandNoInput $command
	}
}


###################################################################
##### ToggleAllWindows{} function 
## NOTE: "nearvalue" must be a global in order to have it's new
##   assignment (in the "-1" case) have the effect of changing
##   the TK label.
proc SetNearvalue { value } {
	global nearvalue

	if {$value == -1} {
		puts "Setting the system near value to default"
		SendCommandNoInput "set near -1"
		set nearvalue [SendCommandGetInput "print near"]
	} elseif {$value <= 0.0} {
		#TODO: make this a popup warning
		puts "Warning, invalid value -- '$value'"
	} else {
		puts "Setting the system near value to $value"
		SendCommandNoInput [format "set near %f" $value]
	}
}


###################################################################
##### ToggleAllWindows{} function 
## NOTE: "farvalue" must be a global in order to have it's new
##   assignment (in the "-1" case) have the effect of changing
##   the TK label.
proc SetFarvalue { value } {
	global farvalue

	if {$value == -1} {
		puts "Setting the system far value to default"
		SendCommandNoInput "set far -1"
		set farvalue [SendCommandGetInput "print far"]
	} elseif {$value <= 0.0} {
		#TODO: make this a popup warning
		puts "Warning, invalid value -- '$value'"
	} else {
		puts "Setting the system far value to $value"
		SendCommandNoInput [format "set far %f" $value]
	}
}


###################################################################
##### MakeProcessInfoFrame{} creates a frame for the display of information
# for a single process
# NOTE: the "eval"s are needed to cause '$format' to be expanded first
# TODO: figure out how to get the "justify" label option to work!
proc MakeProcessInfoFrame {procnum} {
	if {$procnum == "header"} {
		set procframe ".newproc.header"
		set format "-relief ridge -background #EBE -justify center"
	} else {
		set procframe [format ".newproc.proc%d" $procnum]
		set format "-relief ridge -justify right"
	}

	frame $procframe
	eval label $procframe.name  -width 20 $format -text {"Process Name"}
	eval label $procframe.type  -width 10 $format -text {"Type"}
	eval label $procframe.pid   -width 15 $format -text {"PID/TID"}
	eval label $procframe.sync  -width 10 $format -text {"Sync ID"}
	eval label $procframe.count -width 10 $format -text {"Frames"}
	eval label $procframe.fps1  -width 10 $format -text {"FPS-1"}
	eval label $procframe.fps10 -width 10 $format -text {"FPS-10"}
	pack $procframe.name $procframe.type $procframe.pid $procframe.sync $procframe.count $procframe.fps1 $procframe.fps10 -side left

	return $procframe
}


###################################################################
##### HandleProcessInfo{} function takes an input string and parses
# it into the appropriate widget labels
proc HandleProcessInfo { procnum input } {
	set procframe [format ".newproc.proc%d" $procnum]

	set procinfo_list [split $input :]

	set num [lindex $procinfo_list 0]
	if {$num != $procnum} {
		puts "Warning, did not get the expected process information -- expected '$procnum' got '$num'"
		puts "input was '$input'"
		return
	}
	$procframe.name  configure -text [lindex $procinfo_list 2]
	$procframe.type  configure -text [lindex $procinfo_list 3]
	$procframe.pid   configure -text [lindex $procinfo_list 1]
	$procframe.sync  configure -text [lindex $procinfo_list 4]
	$procframe.count configure -text [lindex $procinfo_list 5]
	$procframe.fps1  configure -text [lindex $procinfo_list 6]
	$procframe.fps10 configure -text [lindex $procinfo_list 7]
}


###################################################################
##### ProcUpdate{} function reads the information for the first 4 processes
# and puts that information in a crude line of text form on the TK display
# NOTE: this process can be looped over to continuously update the information.
proc ProcUpdate {} {
	global num_procs

	for { set count 0 } { $count < $num_procs } { incr count } {
		set input [SendCommandGetInput "p proc\[$count\]"]
		HandleProcessInfo $count $input
	}
}


###################################################################
##### MakeUserInfoFrame{} creates a frame for the display of information
# for a single user
# NOTE: the "eval"s are needed to cause '$format' to be expanded first
# TODO: figure out how to get the "justify" label option to work!
proc MakeUserInfoFrame {usernum} {
	if {$usernum == "header"} {
		set userframe ".users.header"
		set format "-relief ridge -background #EBE -justify center"
		set matformat "-justify right"
	} else {
		set userframe [format ".users.user%d" $usernum]
		set format "-relief ridge -justify right"
		set matformat "-justify right -background #FFF"
	}

	frame $userframe
	frame $userframe.mat -relief ridge
	eval label $userframe.name   -width 20 $format -text {"User Name"}
	eval label $userframe.iod    -width 15 $format -text {"IOD"}
	eval label $userframe.eyes   -width 10 $format -text {"Eyes"}
	eval label $userframe.mat.0  -width  8 $matformat -text {"mat0"}
	eval label $userframe.mat.1  -width  8 $matformat -text {"mat1"}
	eval label $userframe.mat.2  -width  8 $matformat -text {"mat2"}
	eval label $userframe.mat.3  -width  8 $matformat -text {"mat3"}
	eval label $userframe.mat.4  -width  8 $matformat -text {"mat4"}
	eval label $userframe.mat.5  -width  8 $matformat -text {"mat5"}
	eval label $userframe.mat.6  -width  8 $matformat -text {"mat6"}
	eval label $userframe.mat.7  -width  8 $matformat -text {"mat7"}
	eval label $userframe.mat.8  -width  8 $matformat -text {"mat8"}
	eval label $userframe.mat.9  -width  8 $matformat -text {"mat9"}
	eval label $userframe.mat.10 -width  8 $matformat -text {"mat10"}
	eval label $userframe.mat.11 -width  8 $matformat -text {"mat11"}
	eval label $userframe.mat.12 -width  8 $matformat -text {"mat12"}
	eval label $userframe.mat.13 -width  8 $matformat -text {"mat13"}
	eval label $userframe.mat.14 -width  8 $matformat -text {"mat14"}
	eval label $userframe.mat.15 -width  8 $matformat -text {"mat15"}
	grid $userframe.name $userframe.iod $userframe.eyes
	grid $userframe.mat -columnspan 4 -column 1
	grid configure $userframe.name -columnspan 2 
	grid configure $userframe.iod -column 2 
	grid configure $userframe.eyes -column 3 

	grid $userframe.mat.0  $userframe.mat.4  $userframe.mat.8  $userframe.mat.12
	grid $userframe.mat.1  $userframe.mat.5  $userframe.mat.9  $userframe.mat.13
	grid $userframe.mat.2  $userframe.mat.6  $userframe.mat.10 $userframe.mat.14
	grid $userframe.mat.3  $userframe.mat.7  $userframe.mat.11 $userframe.mat.15

	return $userframe
}


###################################################################
##### HandleUserInfo{} function takes an input string and parses
# it into the appropriate widget labels
proc HandleUserInfo { usernum input } {
	set userframe [format ".users.user%d" $usernum]

	set userinfo_list [split $input :]

	## First check to see that we actually got user data
	if {"user" != [lindex $userinfo_list 1]} {
		puts "Warning, did not get expected user information"
		puts "input was '$input'"
		return
	}

	set num [lindex $userinfo_list 0]
	if {$num != $usernum} {
		puts "Warning, did not get the expected user information -- expected '$usernum' got '$num'"
		puts "input was '$input'"
		return
	}
	$userframe.name  configure -text [format "User: %s" [lindex $userinfo_list 2]]
	$userframe.eyes  configure -text [lindex $userinfo_list 3]
	$userframe.iod   configure -text [lindex $userinfo_list 4]
	$userframe.mat.0  configure -text [lindex $userinfo_list 5]
	$userframe.mat.1  configure -text [lindex $userinfo_list 6]
	$userframe.mat.2  configure -text [lindex $userinfo_list 7]
	$userframe.mat.3  configure -text [lindex $userinfo_list 8]
	$userframe.mat.4  configure -text [lindex $userinfo_list 9]
	$userframe.mat.5  configure -text [lindex $userinfo_list 10]
	$userframe.mat.6  configure -text [lindex $userinfo_list 11]
	$userframe.mat.7  configure -text [lindex $userinfo_list 12]
	$userframe.mat.8  configure -text [lindex $userinfo_list 13]
	$userframe.mat.9  configure -text [lindex $userinfo_list 14]
	$userframe.mat.10 configure -text [lindex $userinfo_list 15]
	$userframe.mat.11 configure -text [lindex $userinfo_list 16]
	$userframe.mat.12 configure -text [lindex $userinfo_list 17]
	$userframe.mat.13 configure -text [lindex $userinfo_list 18]
	$userframe.mat.14 configure -text [lindex $userinfo_list 19]
	$userframe.mat.15 configure -text [lindex $userinfo_list 20]
}


###################################################################
# Do all the initialization -- unless the socket already exists, in
#   which case the initialization has probably already been done.
if { ![info exists sock] } {

	puts "initializing"
	puts ""

	wm title . "FreeVR Remote Interface"

	#####
	# Connect to the FreeVR application (assumes on this machine)
	set sock [socket localhost 3000]
	fconfigure $sock -blocking 0 -buffering line

	#####
	# Setup the machine-style interface
	# NOTE: we get the input, but we ignore it, since it's just the
	#   old prompt.
	SendCommandGetInput "machine"


	#####
	# Create the TK widgets

	frame .top -relief ridge -borderwidth 4
	label .top.title -text "FreeVR Remote Interface" -font "Helvetica -24 bold" -background #AAA
	label .top.appname -width 50 -text "Application: "
	label .top.authname -width 50 -text "Application author(s): "
	label .top.libname -width 50 -text "Library: "
	label .top.sysname -width 50 -text "System: "
	pack .top.title .top.appname .top.authname .top.libname .top.sysname -fill x -expand yes

	# The ".controls" frame is a row with most of the buttons and sliders
	frame .controls

	## The master buttons (like quit, about, help, etc.)
	frame .controls.main -relief ridge -borderwidth 4
	button .controls.main.quit -text "Quit" -command { exit }
	button .controls.main.help -text "Help" -command { HelpPopup }
	button .controls.main.ui -text "App UI popup" -command { UIPopup}
	button .controls.main.stop -text "stop looping" -command { StopLoop }
	button .controls.main.about -text "About" -command { AboutPopup}
	pack .controls.main.quit .controls.main.help .controls.main.ui .controls.main.stop -fill x -expand yes
	pack .controls.main.about -side bottom

	## The Globle Settings
	frame .controls.globtog -width 10c -height 10c -relief ridge -borderwidth 4
	label .controls.globtog.header -text "Global Settings" -background #EBE
	checkbutton .controls.globtog.showfps -text "Show FPS" -variable toggle_fps -command { ToggleAllWindows "fps" $toggle_fps }
	checkbutton .controls.globtog.showstats -text "Show Stats" -variable toggle_stats -command { ToggleAllWindows "stats" $toggle_stats }
	checkbutton .controls.globtog.showui -text "Show UI" -variable toggle_ui -command { ToggleAllWindows "ui" $toggle_ui }
	checkbutton .controls.globtog.stereo -text "Stereo (NYI)" -variable toggle_stereo -command { ToggleStereo $toggle_stereo }
	checkbutton .controls.globtog.frontwire -text "Wireframe-front (NYI)" -variable toggle_frontwire -command { ToggleRenderMode $toggle_frontwire }
	checkbutton .controls.globtog.backwire -text "Wireframe-back (NYI)" -variable toggle_backwire -command { ToggleRenderMode $toggle_backwire }
	pack .controls.globtog.header -fill x
	pack .controls.globtog.showfps .controls.globtog.showstats .controls.globtog.showui .controls.globtog.stereo .controls.globtog.frontwire .controls.globtog.backwire

	# A Near clipping plane entry
	frame .controls.globtog.near
	label .controls.globtog.near.lab -text "Near:"
	entry .controls.globtog.near.entry -relief sunken -bd 2 -textvariable nearvalue
	pack .controls.globtog.near.lab .controls.globtog.near.entry -side left -padx 1m -pady 2m
	pack .controls.globtog.near
	set nearvalue --
	bind .controls.globtog.near.entry <Return> { SetNearvalue $nearvalue }

	# A Far clipping plane entry
	frame .controls.globtog.far
	label .controls.globtog.far.lab -text "Far:"
	entry .controls.globtog.far.entry -relief sunken -bd 2 -textvariable farvalue
	pack .controls.globtog.far.lab .controls.globtog.far.entry -side left -padx 1m -pady 2m
	pack .controls.globtog.far
	set farvalue --
	bind .controls.globtog.far.entry <Return> { SetFarvalue $farvalue }

  if {0} {
	# A Test Entry (I'm currently just using this to test entry stuff)
	frame .controls.globtog.test
	label .controls.globtog.test.lab -text "test:"
	entry .controls.globtog.test.entry -relief sunken -bd 2 -textvariable test
	set test --
	pack .controls.globtog.test.lab .controls.globtog.test.entry -side left -padx 1m -pady 2m
	pack .controls.globtog.test

	# A Test binding
	bind .controls.globtog.near.entry <Key> {.controls.globtog.test.lab configure -text "keysym = %K"}
  }

	## The loop controls (none yet available)
	frame .controls.loop -relief ridge -borderwidth 4
	label .controls.loop.header -width 20 -anchor w -text "Loop Controls" -background #EBE
	checkbutton .controls.loop.updateprocs -text "Update Procs" -variable update_procs
	checkbutton .controls.loop.updateusers -text "Update Users" -variable update_users
	pack .controls.loop.header .controls.loop.updateprocs .controls.loop.updateusers

	pack .controls.main .controls.globtog .controls.loop -side left -fill y -expand yes


	#####
	# Get the basic system information (and fill in TK widgets)
	## NOTE: this is where num_procs, num_windows & num_users are set
	SystemUpdate

	## The Processes Frame
	#frame .procs -width 10c -height 10c -relief ridge -borderwidth 4
	#label .procs.line1 -width 50 -anchor w -text "line 1" -font "Fixed -20"
	#label .procs.line2 -width 50 -anchor w -text "line 2" -font "Fixed -20"
	#label .procs.line3 -width 50 -anchor w -text "line 3" -font "Fixed -20"
	#label .procs.line4 -width 50 -anchor w -text "line 4" -font "Fixed -20"
	#pack .procs .procs.line1 .procs.line2 .procs.line3 .procs.line4

	## The new Processes Frame
	frame .newproc -width 10c -relief ridge -borderwidth 4
	grid configure [MakeProcessInfoFrame header] -row 0

	for { set count 0 } { $count < $num_procs } { incr count } {
		grid configure [MakeProcessInfoFrame $count] -row [expr $count + 1]
	}


	## The User Frame
	frame .users -width 10c -relief ridge -borderwidth 4
	grid configure [MakeUserInfoFrame 0] -row 0


	#####
	# Do the overall layout
	grid configure .top -row 0 -sticky ew
	grid configure .controls -row 1 -sticky ew
	grid configure .newproc -row 2 -sticky ew
	grid configure .users -row 3 -sticky ew
}


###################################################################
# Get some information about the barriers/locks
set barrier_num [SendCommandGetInput "barrier num"]
puts "$barrier_num barriers"

set lock_num [SendCommandGetInput "lock num"]
puts "$lock_num locks"


set nearvalue [SendCommandGetInput "print near"]
set farvalue [SendCommandGetInput "print far"]

#####
# Loop for a while, updateing the data.
set stoploop_flag 0
while {!$stoploop_flag} {
	if {$update_procs} {
		ProcUpdate
	}

	if {$update_users} {
		set input [SendCommandGetInput "p user\[0\]"]
		HandleUserInfo 0 $input
	}

	update
	after 200
}
puts "Done Looping"

